# -*- coding: utf-8 -*-
# Author: MangoGO (https://blog.csdn.net/silencewx1005)
# Time  : 2024/3/1 21:50
# Desc  :
import time
from os import path

import pandas as pd
import torch
from torch.utils.data import DataLoader
from sklearn.model_selection import train_test_split

from data import DFData, TFIDF, NNDataset, nn_collate_func
from parameters import Parameter
from utils import Log, filter_chinese
from functions import cross_entropy


class NNClassify(torch.nn.Module):
    def __init__(self, in_size, out_size, hidden_size=512):
        super(NNClassify, self).__init__()
        self.model_index = {}
        self.in_size = in_size
        self.out_size = out_size
        self.hidden_size = hidden_size

        self.dense1 = torch.nn.Linear(self.in_size, self.hidden_size)
        self.activation1 = torch.nn.ReLU()
        self.dense2 = torch.nn.Linear(self.hidden_size, self.out_size)
        self.activation2 = torch.nn.Softmax(dim=-1)

    def forward(self, x):
        return self.activation2(self.dense2(self.activation1(self.dense1(x))))

    def save_model(self, save_path, pt_name='pytorch_states.pt',
                   index_name='model_indexes.txt'):
        pt_path = path.join(save_path, pt_name)
        index_path = path.join(save_path, index_name)

        model_all_state = self.module if hasattr(self, 'module') else self
        all_state = model_all_state.state_dict()
        torch.save(all_state, pt_path)

        with open(index_path, mode='w', encoding='utf-8') as f:
            f.write(str(self.model_index).replace("'", '"'))

    def load_model(self, model_path, device):
        update_state_dict = torch.load(model_path, map_location=device)
        model_all_state = self.module if hasattr(self, 'module') else self
        all_state = model_all_state.state_dict()
        all_state.update(update_state_dict)

        self.load_state_dict(all_state)


def test_eval(pred_list, labels):
    acc = int(torch.sum((pred_list - labels == 0).long())) / len(pred_list)

    return acc


def evaluate():
    torch.cuda.empty_cache()

    comment_model.eval()
    with torch.no_grad():
        for batch in eval_dataloader:
            (dt, label) = batch
            dt = dt.to(dev)
            eval_pred = comment_model(dt)
            eval_pred = torch.max(eval_pred, 1)[1]
            acc = test_eval(eval_pred.cpu(), label)
    comment_model.train()

    return acc, eval_pred


def upper_limit(pre_value, not_up_time, curr_value, model_index, pred_res):
    if pre_value is None:
        pre_value = float('-inf')
    if curr_value > pre_value + paras.get('nn_early_stop_inc'):
        not_up_time = 0
        comment_model.model_index = model_index
        comment_model.save_model(paras.get('nn_trained_model_path'))
    else:
        not_up_time += 1

    return curr_value, not_up_time


def train(epochs):
    def eval_show(oth_s=''):
        epoch_s = f'epoch:{epoch - not_up_time + 1: >3}/{epochs: <3},'
        loss_s = f'train loss:{loss: 8f},'
        step_s = f'step:{step + 1: 4}/{batch_amt: <4},' \
            if show_position == 'batch' else ''
        test_s = f'acc:{acc: 8f}' if paras.get('need_test') else ''
        full_s = ' '.join([oth_s, epoch_s, step_s, loss_s, test_s]).strip()
        log.info(full_s)
        return full_s

    def model_index():
        return {'start_min': start_min, 'loss': loss.item(), 'acc': acc} \
            if paras.get('nn_need_test') else {'loss': loss.item()}

    loss, acc, eval_res = torch.Tensor([float('inf')]), 0, 0
    pre_value, not_up_time = None, 0

    epoch_logs = {}
    for epoch in range(epochs):
        # log.info(f'{epoch + 1} epoch for train:')
        torch.cuda.empty_cache()
        comment_model.train()
        for step, batch in enumerate(train_dataloader):
            show_position = 'batch'

            (dt, label) = batch
            dt = dt.to(dev)

            pre_res = comment_model(dt)

            loss = cross_entropy(pre_res.cpu(), label)

            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            if paras.get('nn_show_type') == 'batch' \
                    and (step + 1) % paras.get('nn_show_size') == 0:
                acc, eval_res = evaluate() if paras.get('nn_need_test') \
                    else (None, None)
                eval_show()

        acc, eval_res = evaluate() if paras.get('nn_need_test') \
            else (None, None)
        show_position = 'epoch'
        epoch_logs[epoch + 1] = eval_show('完整训练轮次结果:')

        # 早停处理
        if paras.get('nn_early_stop_target') is None:
            continue
        early_index = {'loss': loss * -1, 'acc': acc}
        curr_value = early_index.get(paras.get('nn_early_stop_target'))

        pre_value, not_up_time = upper_limit(
            pre_value, not_up_time, curr_value,
            model_index(), eval_res)

        if not_up_time >= paras.get('nn_early_stop'):
            print('\n')
            print(f'早停指标：{paras.get("nn_early_stop_target")}')
            print(f'最佳训练轮次: {epoch - not_up_time + 1: 3}, '
                  f'{epoch_logs.get(epoch - not_up_time + 1)}')
            return

    comment_model.model_index = model_index()
    comment_model.save_model(paras.get('nn_trained_model_path'))


if __name__ == '__main__':
    parameter = Parameter()
    paras = parameter.args
    start_min = time.strftime("%Y%m%d%H%M", time.localtime())
    log_path = path.join(paras.get('log_folder'),
                         'log_{}.txt'.format(start_min))
    log = Log(level='debug', file=log_path)
    log.debug('Parameters: {}\n'.format(paras))

    dev = torch.device(paras['device'] if torch.cuda.is_available() else 'cpu')

    train_df = DFData(paras.get('nn_train_data_path')).get_data()

    train_comments = train_df['comment'].apply(
        lambda x: filter_chinese(str(x))).to_list()
    train_tfidf = TFIDF(
        train_comments, cache_path=paras.get('tfidf_path')).sentence_tfidf()
    train_label = train_df['label'].to_list()

    train_tfidf, eval_tfidf, train_label, eval_label = \
        train_test_split(train_tfidf, train_label, test_size=200)

    train_dataset = NNDataset(train_tfidf, train_label)
    train_dataloader = DataLoader(train_dataset,
                                  batch_size=paras.get('nn_train_batch_size'),
                                  shuffle=paras.get('nn_train_data_shuffle'),
                                  collate_fn=nn_collate_func,
                                  drop_last=True)
    batch_amt = len(train_dataloader)

    eval_dataset = NNDataset(eval_tfidf, eval_label)
    eval_dataloader = DataLoader(eval_dataset,
                                 batch_size=len(eval_tfidf),
                                 shuffle=paras.get('train_data_shuffle'),
                                 collate_fn=nn_collate_func,
                                 drop_last=True)

    comment_model = NNClassify(len(train_tfidf[0]),
                               paras.get('nn_model_out_size')).to(dev)

    log.info('开始训练')
    optimizer = torch.optim.AdamW(comment_model.parameters(),
                                  lr=paras.get('nn_learning_rate'))

    train(paras.get('nn_train_epochs'))

    log.info('完成训练')
