# -*- coding: utf-8 -*-
# Author: MangoGO (https://blog.csdn.net/silencewx1005)
# Time  : 2024/3/1 21:50
# Desc  :
import time
from os import path

import torch
from torch.utils.data import DataLoader
from transformers import BertTokenizer

from data import DFData, CommentDataset, comment_collate_func
from model import CustomConfig, CommentModel
from parameters import Parameter
from utils import Log
from functions import cross_entropy


def test_eval(pred_list, labels):
    acc = int(torch.sum((pred_list - labels == 0).long())) / len(pred_list)

    return acc


def evaluate():
    torch.cuda.empty_cache()

    comment_model.eval()
    with torch.no_grad():
        for batch in eval_dataloader:
            (ids, mask, type_ids, label) = batch
            ids, mask, type_ids = ids.to(dev), mask.to(dev), type_ids.to(dev)
            eval_pred = comment_model(ids, mask, type_ids)
            eval_pred = torch.max(eval_pred, 1)[1]
            acc = test_eval(eval_pred.cpu(), label)
    comment_model.train()

    return acc, eval_pred


def upper_limit(pre_value, not_up_time, curr_value, model_index, pred_res):
    if pre_value is None:
        pre_value = float('-inf')
    if curr_value > pre_value + paras.get('early_stop_inc'):
        not_up_time = 0
        comment_model.model_index = model_index
        comment_model.save_model(paras.get('trained_model_path'))

        eval_df['test_res'] = pred_res.cpu()
        eval_df.to_csv(paras['test_res_path'], sep=',', header=False,
                       index=False)
    else:
        not_up_time += 1

    return curr_value, not_up_time


def train(epochs):
    def eval_show(oth_s=''):
        epoch_s = f'epoch:{epoch - not_up_time + 1: 3}/{epochs: <2},'
        loss_s = f'train loss:{loss: 8f},'
        step_s = f'step:{step + 1: 4}/{batch_amt: <4},' \
            if show_position == 'batch' else ''
        test_s = f'acc:{acc: 8f}' if paras.get('need_test') else ''
        full_s = ' '.join([oth_s, epoch_s, step_s, loss_s, test_s]).strip()
        log.info(full_s)
        return full_s

    def model_index():
        return {'start_min': start_min, 'loss': loss.item(), 'acc': acc} \
            if paras.get('need_test') else {'loss': loss.item()}

    loss, acc, eval_res = torch.Tensor([float('inf')]), 0, 0
    pre_value, not_up_time = None, 0

    epoch_logs = {}
    for epoch in range(epochs):
        log.info(f'{epoch + 1} epoch for train:')
        torch.cuda.empty_cache()
        comment_model.train()
        for step, batch in enumerate(train_dataloader):
            show_position = 'batch'

            (ids, mask, type_ids, label) = batch
            ids, mask, type_ids = ids.to(dev), mask.to(dev), type_ids.to(dev)

            pre_res = comment_model(ids, mask, type_ids)

            loss = cross_entropy(pre_res.cpu(), label)

            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            if paras.get('show_type') == 'batch' \
                    and (step + 1) % paras.get('show_size') == 0:
                acc, eval_res = evaluate() if paras.get('need_test') \
                    else (None, None)
                eval_show()

        acc, eval_res = evaluate() if paras.get('need_test') \
            else (None, None)
        show_position = 'epoch'
        epoch_logs[epoch + 1] = eval_show('完整训练轮次结果:')

        # 早停处理
        if paras.get('early_stop_target') is None:
            continue
        early_index = {'loss': loss * -1, 'acc': acc}
        curr_value = early_index.get(paras.get('early_stop_target'))

        pre_value, not_up_time = upper_limit(
            pre_value, not_up_time, curr_value,
            model_index(), eval_res)

        if not_up_time >= paras.get('early_stop'):
            print('\n')
            print(f'早停指标：{paras.get("early_stop_target")}')
            print(f'最佳训练轮次: {epoch - not_up_time + 1: 3}, '
                  f'{epoch_logs.get(epoch - not_up_time + 1)}')
            return

    comment_model.model_index = model_index()
    comment_model.save_model(paras.get('trained_model_path'))

    eval_df['test_res'] = eval_res.cpu()
    eval_df.to_csv(paras.get('test_res_path'), sep=',',
                   header=False, index=False)


if __name__ == '__main__':
    parameter = Parameter()
    paras = parameter.args
    start_min = time.strftime("%Y%m%d%H%M", time.localtime())
    log_path = path.join(paras.get('log_folder'),
                         'log_{}.txt'.format(start_min))
    log = Log(level='debug', file=log_path)
    log.debug('Parameters: {}\n'.format(paras))

    dev = torch.device(paras.get('device') if torch.cuda.is_available()
                       else 'cpu')

    model_config = CustomConfig.from_pretrained_with_define(
        paras.get('pre_model_path'), paras)
    tokenizer = BertTokenizer.from_pretrained(paras.get('pre_model_path'))

    train_df = DFData(paras.get('train_data_path')).get_data()
    eval_df = DFData(paras.get('eval_data_path')).get_data()
    # test_df = DFData(paras.get('test_data_path')).get_data()

    train_dataset = CommentDataset(train_df, tokenizer, paras.get('max_len'))
    train_dataloader = DataLoader(train_dataset,
                                  batch_size=paras.get('train_batch_size'),
                                  shuffle=paras.get('train_data_shuffle'),
                                  collate_fn=comment_collate_func,
                                  drop_last=True)
    batch_amt = len(train_dataloader)

    eval_dataset = CommentDataset(eval_df, tokenizer, paras.get('max_len'))
    eval_dataloader = DataLoader(eval_dataset,
                                 batch_size=eval_df.shape[0],
                                 shuffle=paras.get('train_data_shuffle'),
                                 collate_fn=comment_collate_func,
                                 drop_last=True)

    comment_model = CommentModel(paras.get('pre_model_path'), model_config,
                                 paras.get('model_out_size')).to(dev)

    log.info('开始训练')
    optimizer = torch.optim.AdamW(comment_model.parameters(),
                                  lr=paras.get('learning_rate'))

    train(paras.get('train_epochs'))

    log.info('完成训练')
