# -*- coding: utf-8 -*-
# Author: MangoGO (https://blog.csdn.net/silencewx1005)
# Time  : 2024/2/19 21:00
# Desc  :
import json
import math
import re
from os import path
from secrets import SystemRandom

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset
import jieba

from utils import path2files, get_json


def secret_random(rate):
    return SystemRandom().randrange(0, 1000000, 1) > rate * 1000000


class DFData:
    """Read original_data and return DataFrame

    According to the original_data folder or original_data file name,
    combined with the original_data configuration file, read the original_data.
    If the path is a folder, read all original_data files in the folder
    and merge the output.
    """
    def __init__(self, data_path=None, config_path=None, engine=None):
        self.data_path = data_path
        self.config_path = config_path
        self.engine = engine

        self.column_sep = None
        self.new_columns = None
        self.data_fold = None
        self.data_names = None
        self.table_name = None
        self.sep = ','
        self.header_num = None
        self.columns = None
        self.columns_idx = None
        self.shape = None
        self.primary_keys = None
        self.dict_columns = None
        self.tensor_columns = None
        if self.config_path is not None:
            self.get_table_info()

        self.df = None
        self.split_flag = False

    @staticmethod
    def rate_check(rate):
        if 0 <= rate <= 1:
            return True
        else:
            return False

    @staticmethod
    def join_element(e1, e2, join_type):
        if join_type == 'list':
            return [e1, e2]
        if join_type == 'tuple':
            return tuple([e1, e2])
        if isinstance(join_type, str):
            return str(e1) + join_type + str(e2)

        raise ValueError(f'join_type期待字符串类型参数, 但得到{join_type}!')

    @staticmethod
    def str2tensor(string, del_pattern, sel_pattern):
        del_reg = re.compile(del_pattern)
        sel_reg = re.compile(sel_pattern)

        if isinstance(string, torch.Tensor):
            return string

        try:
            string = del_reg.sub('', string)
            digits = np.array([[float(ele) for ele in sel_reg.findall(string)]])
            return torch.Tensor(digits)
        except ValueError as e:
            print(f'Error converting string to floating-point type! '
                  f'The string to be converted is: {string}')
            raise e
        except TypeError as e:
            print(f'Error converting string to floating-point type! '
                  f'The string to be converted is: {string}')
            raise e

    def df_str2tensor(self):
        for c, [del_pattern, sel_pattern] in self.tensor_columns.items():
            self.df[c] = self.df[c].apply(
                lambda x: self.str2tensor(x, del_pattern, sel_pattern))

    def df_str2dict(self):
        for c in self.dict_columns:
            self.df[c] = self.df[c].apply(lambda x: json.loads(
                '{}' if x is np.nan else x.replace("'", '"')))

    def get_table_info(self):
        try:
            if not path.exists(self.config_path):
                raise FileNotFoundError
        except FileNotFoundError:
            print(f'未找到文件: {self.config_path}')
        columns_json = get_json(self.config_path)
        self.table_name = columns_json.get('table_name')
        self.sep = columns_json.get('sep')
        self.column_sep = columns_json.get('column_sep')
        self.new_columns = columns_json.get('new_columns')
        self.header_num = columns_json.get('header')
        self.primary_keys = columns_json.get('primary_keys')
        self.dict_columns = columns_json.get('dict_columns')
        self.tensor_columns = columns_json.get('tensor_columns')
        orig_cols = columns_json.get('orig_cols')
        not_cols = columns_json.get('not_cols')

        if not not_cols:
            self.columns = orig_cols
            self.columns_idx = None
        else:
            self.columns = []
            self.columns_idx = []
            for i, c in enumerate(orig_cols):
                if c not in not_cols:
                    self.columns.append(c)
                    self.columns_idx.append(i)

    def read_datafile(self, filename, rate=None):
        if rate is not None and not self.rate_check(rate):
            raise ValueError(f'Data sampling rate {rate} is not legal!')

        file_type = filename.split('.')[-1]
        if file_type in ('txt', 'csv'):
            df = pd.read_csv(filename, sep=self.sep, encoding='utf-8',
                             names=self.columns, usecols=self.columns_idx,
                             skiprows=lambda x: secret_random(rate)
                             if rate is not None else None,
                             engine=self.engine)
        elif file_type in ('xls', 'xlsx', 'xlsm', 'xlsb'):
            df = pd.read_excel(filename, header=self.header_num,
                               engine=self.engine, names=self.columns,
                               usecols=self.columns_idx,
                               skiprows=lambda x: secret_random(rate)
                               if rate is not None else None)
        else:
            df = pd.DataFrame()

        return df

    def get_data(self, rate=None, return_df=True):
        if self.df is not None and rate is None:
            return self.df

        files = path2files(self.data_path)
        print('数据文件个数: {}'.format(len(files)))

        df = pd.DataFrame()
        for i, file in enumerate(files):
            print('[{}] {}'.format(i, file), end=' ')
            tmp_df = self.read_datafile(file, rate=rate)
            print(tmp_df.shape)
            df = pd.concat([df, tmp_df], axis=0, ignore_index=True)

        print('获取数据量为: {}'.format(df.shape))
        self.df = df
        self.shape = self.df.shape

        if return_df is True:
            return df
        else:
            return None

    def column_split(self):
        if self.df is None:
            raise RuntimeError('执行列分裂之前必须先执行get_data()获取数据!')
        for column, info in self.column_sep.items():
            sep, new_col, keep = info['sep'], info['new_col'], info['keep']
            tmp_series = self.df[column].apply(
                lambda x: [item.strip() for item in str(x).split(sep)
                           if item.strip() != ''] if str(x) != '' else []
            )
            if keep is False:
                self.df.drop(columns=column, inplace=True)
            self.df[new_col] = tmp_series

        self.split_flag = True

    def column_explode(self):
        if not self.split_flag:
            raise RuntimeError('执行列爆炸之前必须先执行列分裂命令column_split()!')
        for _, info in self.column_sep.items():
            self.df = self.df.explode(info['new_col'], ignore_index=True)

    def join_cols(self, series, join_type):
        list1, list2 = series
        if isinstance(list1, str):
            list1 = [list1]
        if isinstance(list2, str):
            list2 = [list2]

        res = []
        for i in list1:
            for j in list2:
                res.append(self.join_element(i, j, join_type))

        return res

    def new_column_create(self):
        if self.df is None:
            raise RuntimeError('执行列创建之前必须先执行get_data()获取数据!')
        for column, info in self.new_columns.items():
            join_cols = info['join_cols']
            if len(join_cols) != 2:
                raise ValueError('待拼接字段数量不为2!')
            join_type = info['join_type']
            tmp_df = self.df[join_cols]

            self.df[column] = tmp_df.apply(
                lambda x: self.join_cols(x, join_type), axis=1)

    def unique_df(self, primary_keys=None, replace_df=True, return_df=False):
        if self.df is None:
            raise RuntimeError('执行去重之前必须先执行get_data()获取数据!')
        if primary_keys is None and self.primary_keys is None:
            raise ValueError('未指定去重字段时按DataFrame主键进行去重，两者皆未指定!')
        if primary_keys is None:
            primary_keys = self.primary_keys

        tmp_df = self.df.drop_duplicates(primary_keys, ignore_index=True)
        if replace_df:
            self.df = tmp_df
        if return_df:
            return tmp_df
        else:
            return None

    def sorted_df(self, sorted_columns=None, replace_df=True, return_df=False):
        if self.df is None:
            raise RuntimeError('执行排序之前必须先执行get_data()获取数据!')
        if sorted_columns is None and self.primary_keys is None:
            raise ValueError('未指定排序字段时按DataFrame主键进行排序, 两者皆未指定!')
        if sorted_columns is None:
            sorted_columns = self.primary_keys

        tmp_df = self.df.sort_values(sorted_columns, ignore_index=True)
        if replace_df:
            self.df = tmp_df
        if return_df:
            return tmp_df
        else:
            return None

    def split_data(self, eval_amt=None, eval_rate=None, test_amt=None,
                   test_rate=None, random_state=None):
        if self.df is None:
            raise RuntimeError('执行数据分割之前必须先执行get_data()获取数据!')
        if eval_amt is not None and eval_rate is not None:
            raise ValueError('eval_amt,eval_rate不能同时被指定!')
        if test_amt is not None and test_rate is not None:
            raise ValueError('test_amt,test_rate不能同时被指定!')
        if eval_rate is not None and not self.rate_check(eval_rate):
            raise ValueError(f'验证集比例{eval_rate}不合法!')
        if test_rate is not None and not self.rate_check(test_rate):
            raise ValueError(f'测试集比例{test_rate}不合法!')

        train_df = self.df
        eval_df, test_df = None, None
        if eval_amt is None and eval_rate is not None:
            eval_amt = int(self.shape[0] * eval_rate)
        if test_amt is None and test_rate is not None:
            test_amt = int(self.shape[0] * test_rate)

        if eval_amt is None and test_amt is None:
            raise ValueError('没有验证集或训练集需要分割!')

        if eval_amt is not None:
            eval_df = train_df.sample(eval_amt, random_state=random_state)
            train_df = train_df[~self.df.index.isin(eval_df.index)]
            eval_df.reset_index(drop=True, inplace=True)
            train_df.reset_index(drop=True, inplace=True)
        if test_amt is not None:
            test_df = train_df.sample(test_amt, random_state=random_state)
            train_df = train_df[~train_df.index.isin(test_df.index)]
            test_df.reset_index(drop=True, inplace=True)
            train_df.reset_index(drop=True, inplace=True)

        return train_df, eval_df, test_df

    def df_sample(self, rate, random_state=None):
        if self.df is None:
            raise RuntimeError('执行数据分割之前必须先执行get_data()获取数据!')
        tmp_df = self.df.sample(frac=rate, random_state=random_state)
        tmp_df.reset_index(drop=True, inplace=True)

        return tmp_df

    def delete_rows(self, conditions, replace_df=True, return_df=False):
        if self.df is None:
            raise RuntimeError('执行数据删除之前必须先执行get_data()获取数据!')

        row_nums = row_index_selector(self.df, conditions)
        tmp_df = self.df.drop(index=row_nums).reset_index(drop=True)

        if replace_df:
            self.df = tmp_df
        if return_df:
            return tmp_df
        else:
            return None

    def save(self, out_path, header=False, rate=None):
        if self.df is None or rate is not None:
            df = self.get_data(rate)
        else:
            df = self.df
        file_type = out_path.split('.')[-1]

        if file_type in ('txt', 'csv'):
            df.to_csv(out_path, sep=self.sep, header=header, index=False,
                      encoding='utf-8')
        elif file_type in ('xls', 'xlsx', 'xlsm', 'xlsb'):
            df.to_excel(out_path, index=False, engine=self.engine)
        else:
            raise ValueError(f'输出文件格式错误！输出文件名称：{out_path}')

    def __len__(self):
        if self.shape is None:
            return self.get_data().shape[0]
        return self.shape[0]


def row_index_selector(df, conditions_info):
    """根据条件从DataFrame中筛选数据

    :param df: 待筛选的DataFrame
    :param conditions_info: 字典，{字段：[筛选值]}，可使用多个字段筛选
    :return: 满足条件的数据的行索引
    """
    conditions = np.array([True] * df.shape[0])
    for k, vs in conditions_info.items():
        tmp_conditions = df[k].apply(lambda x: x in vs)
        conditions = conditions & tmp_conditions

    row_nums = df[conditions].index.tolist()

    if not row_nums:
        raise ValueError(f'按照条件{conditions_info}未找到数据!')

    return row_nums


def auto_padding(seq, max_len, pad=0, device=torch.device('cpu')):
    """序列或矩阵最后维度填充"""
    def pad_list(ls):
        if isinstance(ls[0], int):
            ls += [pad] * (max_len - len(ls))
        else:
            for i, ele in enumerate(ls):
                ls[i] = pad_list(ele)
        return ls

    if isinstance(seq, torch.Tensor):
        shape = list(seq.shape)
        shape[-1] = max_len - shape[-1]
        res = torch.cat([seq, torch.zeros(shape).long().to(device) + pad],
                        dim=-1)
    elif isinstance(seq, list):
        res = pad_list(seq)
    else:
        raise TypeError('待填充的序列或矩阵不是torch.Tensor或list格式!')

    return res


class CommentDataset(Dataset):
    """将格式为DataFrame的评论数据封装成Dataset，迭代输出评论文本的token ids"""
    def __init__(self, df, tokenizer, max_len=512):
        self.df = df
        self.tokenizer = tokenizer
        self.max_len = max_len

    def ids_with_label(self, series):
        comment = series.get('comment')
        label = series.get('label')

        ids = self.tokenizer.convert_tokens_to_ids(list(comment))
        ids = [self.tokenizer.cls_token_id] + ids[:self.max_len-2] + \
              [self.tokenizer.sep_token_id]
        mask = [1] * len(ids)
        type_ids = [0] * len(ids)

        return {'ids': ids, 'mask': mask, 'type_ids': type_ids, 'label': label}

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        return self.ids_with_label(self.df.loc[idx, :])


def comment_collate_func(lines):
    """Dataloader batch内数据长度补齐"""
    ids, mask, type_ids, label = [], [], [], []
    batch_max_len = 0
    for line in lines:
        batch_max_len = max(batch_max_len, len(line.get('ids')))

        ids.append(line.get('ids'))
        mask.append(line.get('mask'))
        type_ids.append(line.get('type_ids'))
        label.append(line.get('label'))

    ids = auto_padding(ids, batch_max_len)
    mask = auto_padding(mask, batch_max_len)
    type_ids = auto_padding(type_ids, batch_max_len)

    ids = torch.LongTensor(ids)
    mask = torch.LongTensor(mask)
    type_ids = torch.LongTensor(type_ids)
    if label[0] is not None:
        label = torch.LongTensor(label)

    return ids, mask, type_ids, label


class TFIDF:
    def __init__(self, documents, cache_path=None, refresh=False):
        self.document_list = documents
        self.cache_path = cache_path
        self.refresh = refresh

        self.word_set = set([word for doc in self.document_list for word in
                             jieba.lcut(doc, cut_all=False)])

        self.doc_dict_list = []
        self.doc_word_num()

        self.tf_list = []
        self.compute_tf()

        self.idf_dict = dict.fromkeys(self.doc_dict_list[0], .0)
        self.compute_idf()

        self.tf_idf_list = []

    def doc_word_num(self):
        """次数统计"""
        for doc in self.document_list:
            word_dict = dict.fromkeys(self.word_set, 0)
            bow = jieba.lcut(doc, cut_all=False)
            for word in bow:
                word_dict[word] += 1
            self.doc_dict_list.append(word_dict)

    def compute_tf(self):
        """计算词频TF"""
        for word_dict, document in zip(self.doc_dict_list, self.document_list):
            tf_dict = {}
            n_bow_count = len(jieba.lcut(document, cut_all=False))
            for word, count in word_dict.items():
                tf_dict[word] = count / (n_bow_count + 1e-9)
            self.tf_list.append(tf_dict)

    def compute_idf(self):
        """计算逆文档频率IDF.用一个字典对象保存idf结果，每个词作为key，初始值为0
        """
        n = len(self.document_list)

        for word_dict in self.doc_dict_list:
            for word, count in word_dict.items():
                if count > 0:
                    self.idf_dict[word] += 1

        for word, ni in self.idf_dict.items():
            self.idf_dict[word] = math.log10((n + 1) / (ni + 1))

    @staticmethod
    def compute_tfidf(tf_dict, idf_dict):
        """计算TF-IDF"""
        tfidf = {}
        for word, tf_val in tf_dict.items():
            tfidf[word] = tf_val * idf_dict[word]
        return tfidf

    def sentence_tfidf(self, return_type='list'):
        tf_idf_list = []
        if self.cache_path is not None and path.exists(self.cache_path) \
                and not self.refresh:
            with open(self.cache_path, mode='r', encoding='utf-8') as f:
                for line in f.readlines():
                    tf_idf_list.append(json.loads(line))
        else:
            for tf in self.tf_list:
                tfidf = self.compute_tfidf(tf, self.idf_dict)
                tf_idf_list.append(tfidf)
            if self.cache_path is not None:
                with open(self.cache_path, mode='w', encoding='utf-8') as f:
                    for line in tf_idf_list:
                        f.write(str(line).replace("'", '"') + '\n')

        if return_type == 'list':
            pass
        elif return_type == 'dataframe':
            tf_idf_list = pd.DataFrame(tf_idf_list)
        elif return_type == 'pt':
            tf_idf_list = pd.DataFrame(tf_idf_list)
            tf_idf_list = torch.Tensor(tf_idf_list.values)
        else:
            raise ValueError(f'tf-idf return type error: {return_type}')

        self.tf_idf_list = tf_idf_list

        return tf_idf_list


class NNDataset(Dataset):
    """将格式为DataFrame的评论数据封装成Dataset，迭代输出评论文本的token ids"""
    def __init__(self, dl, labels):
        self.dl = dl
        self.labels = labels

    def __len__(self):
        return len(self.dl)

    def __getitem__(self, idx):
        return {'dt': list(self.dl[idx].values()), 'label': self.labels[idx]}


def nn_collate_func(lines):
    dt, labels = [], []
    for line in lines:
        dt.append(line.get('dt'))
        labels.append(line.get('label'))

    dt = torch.Tensor(dt)
    if labels[0] is not None:
        labels = torch.LongTensor(labels)

    return dt, labels
