#!/usr/bin/env python
# coding: utf-8
"""

"""
import torch
from scipy.stats import spearmanr as SP
from torch.nn import functional as fx


def correlation_coefficient(nums_1, nums_2):
    res = SP(nums_1.cpu(), nums_2.cpu())
    return res.correlation, res.pvalue


def binary_eval(pred_nums, labeled_nums, positive_baseline=0.5):
    pred_nums = (pred_nums > positive_baseline).long()
    TP = int(torch.sum((pred_nums + labeled_nums == 2).long()))
    TN = int(torch.sum((pred_nums + labeled_nums == 0).long()))
    FP = int(torch.sum((pred_nums - labeled_nums == 1).long()))
    FN = int(torch.sum((pred_nums - labeled_nums == -1).long()))

    acc = (TP + TN) / (len(pred_nums) + 1e-10)
    precision = TP / (TP + FP + 1e-10)
    recall = TP / (TP + FN + 1e-10)
    f1 = 2 * precision * recall / (precision + recall + 1e-10)

    return {'acc': acc, 'precision': precision, 'recall': recall, 'f1': f1}


def multi_classify_eval(pred_type, label_type, key_pre=''):
    """比较预测序列与标记序列，计算准确率、精确率、召回率、f1

    准确率acc: 预测正确的token个数/序列长度
    精确率precision: 预测非0的token中，预测正确的token个数/标预测非0的token个数
    召回率recall: 标记非0的token中，预测正确的token个数/标记非0的token个数
    f1: 2 * precision * recall / (precision + recall)
    :param pred_type: 预测结果序列，一维张量
    :param label_type: 标记序列， 一维张量
    :param key_pre: 输出结果中key前缀
    :return: 准确率、精确率、召回率、f1
    """
    equal_series = pred_type == label_type
    pred_positive = pred_type > 0
    label_positive = label_type > 0

    pred_pos_num = int(torch.sum(pred_positive.long()))
    label_pos_num = int(torch.sum(label_positive.long()))
    pred_pos_eq_num = int(torch.sum((pred_positive & equal_series).long()))
    label_pos_eq_num = int(torch.sum((label_positive & equal_series).long()))

    acc = int(torch.sum(equal_series.long())) / (equal_series.shape[0] + 1e-10)
    precision = pred_pos_eq_num / (pred_pos_num + 1e-10)
    recall = label_pos_eq_num / (label_pos_num + 1e-10)
    f1 = 2 * precision * recall / (precision + recall + 1e-10)

    return {key_pre + 'acc': acc, key_pre + 'precision': precision,
            key_pre + 'recall': recall, key_pre + 'f1': f1}


def cross_entropy(pred_probs, y_label, mask=None):
    """多分类交叉熵损失函数

    :param pred_probs: 多分类预测各分类概率结果, 经softmax处理
    :param y_label: 实际标签
    :param mask: 只关注有效结果
    :return: 损失
    """
    pred_log = torch.log(pred_probs + 1e-10)  # 增加稳定系数

    if len(pred_probs.shape) == 2:
        pass
    elif len(pred_probs.shape) == 3:
        pred_log = pred_log.transpose(1, 2)
    else:
        raise ValueError('只支持2维或3维张量计算损失!')
    loss = fx.nll_loss(pred_log, y_label, reduction='none')

    if mask is None:
        loss = loss.mean()
    else:
        loss = torch.sum(loss * mask) / (torch.sum(mask) + 1e-10)

    return loss
