#!/usr/bin/env python
# coding: utf-8
"""算法模型搭建

"""
import io
import os

import torch
from torch import nn
from transformers import BertModel, BertConfig


class BertBase(nn.Module):
    def __init__(self, pre_model_path, config, num_embedding=None):
        super(BertBase, self).__init__()
        self.bert = BertModel.from_pretrained(pre_model_path, config=config)

        if num_embedding is not None:
            self.resize_token_embedding(num_embedding)
        self.model_index = {}
        self.define_layers = []

    def resize_token_embedding(self, num_embedding):
        self.bert.resize_token_embeddings(num_embedding)

    def save_model(self, save_path, pt_name='pytorch_states.pt', bert_name=None,
                   index_name='model_indexes.txt'):
        pt_path = os.path.join(save_path, pt_name)
        index_path = os.path.join(save_path, index_name)

        model_all_state = self.module if hasattr(self, 'module') else self
        all_state = model_all_state.state_dict()
        save_state = {}
        for param_tensor in all_state:
            layer_name = param_tensor.split('.')[0]
            if layer_name in self.define_layers:
                save_state.update({param_tensor: all_state[param_tensor]})
        torch.save(save_state, pt_path)

        orig_bert_name = os.path.join(save_path, 'pytorch_model.bin')
        self.bert.save_pretrained(save_path)
        if bert_name is not None and bert_name != 'pytorch_model.bin':
            aim_bert_name = os.path.join(save_path, bert_name)
            os.rename(orig_bert_name, aim_bert_name)

        with open(index_path, mode='w', encoding='utf-8') as f:
            f.write(str(self.model_index).replace("'", '"'))

    def load_model(self, model_path, device):
        update_state_dict = torch.load(model_path, map_location=device)
        model_all_state = self.module if hasattr(self, 'module') else self
        all_state = model_all_state.state_dict()
        all_state.update(update_state_dict)

        self.load_state_dict(all_state)


class CustomConfig:
    paras = None
    config = None

    def __init__(self):
        pass

    @classmethod
    def from_pretrained_with_define(cls, pre_model_path, define_paras=None):
        cls.config = BertConfig.from_pretrained(pre_model_path)
        cls.paras = define_paras

        for k, _ in cls.config.__dict__.items():
            if cls.paras is None or cls.paras.get(k) is None:
                continue
            cls.config.__dict__[k] = cls.paras.get(k)

        return cls.config


class CommentModel(BertBase):
    def __init__(self, pre_model_path, config, out_size, num_embedding=None):
        super(CommentModel, self).__init__(pre_model_path, config,
                                           num_embedding)
        self.dense = nn.Linear(config.hidden_size, out_size)
        self.activation = nn.Softmax(dim=-1)
        self.define_layers = ['dense']

    def forward(self, input_ids, attention_mask, token_type_ids):
        bert_output = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids
        )
        bert_output = bert_output.pooler_output

        output = self.activation(self.dense(bert_output))
        return output
