# -*- coding: utf-8 -*-
# Author: MangoGO (https://blog.csdn.net/silencewx1005)
# Time  : 2024/1/24 22:18
# Desc  : A variety of basic tool functions

import json
import logging
import re
import sys
import os
import shutil
import time
from os import path

import requests


class Log:
    """Print colorless logs based on different log level"""
    font_color = {
        'black': '30',
        'red': '31',
        'green': '32',
        'yellow': '33',
        'blue': '34',
        'carmine': '35',
        'cyan': '36',
        'white': '37'
    }
    show_level = {
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'debug': logging.DEBUG
    }

    def __init__(self, name='log', file=None, level='warning',
                 formatted='%(asctime)s %(message)s',
                 datefmt='%Y-%m-%d %H:%M:%S',
                 info_color=None, warning_color=None, error_color=None,
                 debug_color=None):
        self.name = name
        self.file = file
        self.format = formatted
        self.datefmt = datefmt
        level = self.show_level.get(level)
        self.level = level if level is not None \
            else self.show_level.get('warning')
        self.info_color = info_color
        self.warning_color = warning_color
        self.error_color = error_color
        self.debug_color = debug_color

        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(self.level)
        self._show_config()

    def info(self, message):
        color = self.font_color.get(self.info_color)
        color_msg = message if color is None \
            else "\033[0;{}m{}\033[0m".format(color, message)
        self.logger.info(color_msg)

    def warning(self, message):
        color = self.font_color.get(self.warning_color)
        color_msg = message if color is None \
            else "\033[0;{}m{}\033[0m".format(color, message)
        self.logger.warning(color_msg)

    def error(self, message):
        color = self.font_color.get(self.error_color)
        color_msg = message if color is None \
            else "\033[0;{}m{}\033[0m".format(color, message)
        self.logger.error(color_msg)

    def debug(self, message):
        color = self.font_color.get(self.debug_color)
        color_msg = message if color is None \
            else "\033[0;{}m{}\033[0m".format(color, message)
        self.logger.debug(color_msg)

    def _show_config(self):
        formatting = logging.Formatter(fmt=self.format, datefmt=self.datefmt)

        sh = logging.StreamHandler(stream=sys.stdout)
        sh.setFormatter(formatting)
        self.logger.addHandler(sh)

        if self.file is not None:
            fh = logging.FileHandler(self.file, encoding="UTF-8")
            fh.setFormatter(formatting)
            self.logger.addHandler(fh)


def flat_key(layer):
    if len(layer) == 1:
        return layer[0]
    else:
        ls = [f'[{k}]' for k in layer[1:]]
        return str(layer[0]) + ''.join(ls)


def flat_dict(input_dict):
    if not isinstance(input_dict, dict):
        raise TypeError(f'input_dict should be a dict，'
                        f'but got {type(input_dict)}')

    def cycle_flat_dict(pre_layer, value):
        result = {}
        for k, v in value.items():
            layer = pre_layer[:]
            layer.append(k)
            if isinstance(v, dict):
                result.update(cycle_flat_dict(layer, v))
            else:
                result[flat_key(layer)] = v
        return result

    return cycle_flat_dict([], input_dict)


def print_eval_res(res_dict, title=False, sep=' | '):
    """Print a dict without any quotation mark, supporting multilevel dict

    :param res_dict: The dict to be printed
    :param title: Whether to include column names when printing
    :param sep: The separator between columns
    :return:
    """
    one_level_dict = flat_dict(res_dict)
    ks, vs = list(one_level_dict.keys()), list(one_level_dict.values())
    if title:
        Log().info(sep.join([str(i) for i in ks]))
    Log().info(sep.join([str(i) for i in vs]))


def get_json(file_path):
    with open(file_path, mode='r', encoding='utf-8') as jsonfile:
        json_dict = json.load(jsonfile)
    return json_dict


def delete_folder(dpath):
    if path.exists(dpath):
        Log().info(f'Empty the folder {dpath}')
        shutil.rmtree(dpath)


def reset_folder(dpath):
    if path.exists(dpath):
        Log().info(f'Empty the folder {dpath}')
        shutil.rmtree(dpath)
    os.makedirs(dpath)


def str2bool(v):
    if v.lower() in ('no', 'false', 'f', '0', 0):
        return False
    else:
        return True


def str2list(v, defined_type=True):
    v = v.replace('[', '').replace(']', '').replace('"', '').replace("'", '')
    res = [i.strip() for i in v.split(',')]
    if defined_type:
        ele_type = str2type(res[0])
        res = res[1:]
    else:
        ele_type = str2type('str')
    res = [ele_type(i) for i in res]

    return res


def str2type(v):
    if v == 'int':
        return int
    elif v == 'float':
        return float
    elif v == 'bool':
        return str2bool
    elif v == 'list':
        return str2list
    elif v == 'tuple':
        return tuple
    else:
        return str


def dict2type(dtypes):
    for k, v in dtypes.items():
        dtypes[k] = str2type(v)

    return dtypes


def is_path_or_dir(string):
    if path.isfile(string):
        return 'file'
    elif path.isdir(string):
        return 'dir'
    else:
        return ''


def path2files(dpath, return_type='all'):
    if return_type == 'all':
        return_types = ['file', 'dir']
    elif return_type in ('file', 'dir'):
        return_types = [return_type]
    else:
        raise ValueError(f'return_type must be "all", "file" or "dir", '
                         f'but got {return_type}')
    files = []
    if path.isfile(dpath):
        files = [dpath]
    elif path.isdir(dpath):
        files = [path.join(dpath, f) for f in os.listdir(dpath)]

    files = [file for file in files if is_path_or_dir(file) in return_types]
    return files


def is_float(v):
    if v.count('.') == 1 and (v.count('-') == 0 or v[0] == '-'):
        v = v.replace('.', '').replace('-', '')
        return v.isdigit()
    return False


def files_chooser(files, task_amt, task_id):
    if task_amt is None or task_id is None:
        return files
    file_amt = len(files)
    consult = file_amt // task_amt
    reminder = file_amt % task_amt

    if reminder == 0:
        min_size, max_size = consult, consult
    else:
        min_size, max_size = consult, consult + 1

    start_idx = [0]
    for idx in range(task_amt):
        batch_size = max_size if idx < reminder else min_size
        start_idx.append(start_idx[-1] + batch_size)

    return files[start_idx[task_id]: start_idx[task_id + 1]]


def filter_chinese(s):
    pattern = re.compile(r'[\u4e00-\u9fa5]+')
    chinese = pattern.findall(s)

    return ' '.join(chinese)
